#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

__all__ = ['AnalyzeThread',
           'PluginReportReceiver',
           'SETroubleshootDatabase',
           'analyze_logfile',
          ]

PROGNAME="setroubleshoot"
import gettext
gettext.bindtextdomain(PROGNAME, "/usr/share/locale")
gettext.textdomain(PROGNAME)
try:
    gettext.install(PROGNAME, "/usr/share/locale", 1)
except IOError:
    import __builtin__
    __builtin__.__dict__['_'] = unicode

from setroubleshoot.avc_audit import *
from setroubleshoot.config import cfg
from setroubleshoot.errcode import *
from setroubleshoot.log import *
from setroubleshoot.signature import *
from setroubleshoot.util import *

import errno
import os
import Queue
import signal
from stat import *
import re
import sys
import time
import threading
from types import *

#------------------------------------------------------------------------------

class PluginStatistics(object):
    def __init__(self, plugin):
        self.name = plugin.analysis_id
        self.analyze_start_time = None
        self.analyze_end_time = None
        self.analyze_elapsed_time = None
        self.report_start_time = None
        self.report_end_time = None
        self.report_elapsed_time = None

    def __str__(self):
        analyze_elapsed_time = format_elapsed_time(self.analyze_elapsed_time)

        if self.report_elapsed_time is None:
            return "%s: %s elapsed" % (self.name, analyze_elapsed_time)
        else:
            total_elapsed_time = format_elapsed_time(self.report_end_time-self.analyze_start_time)
            report_elapsed_time = format_elapsed_time(self.report_elapsed_time)

            return "%s: %s elapsed, %s analyze elapsed, %s report elapsed" % \
                   (self.name, total_elapsed_time, analyze_elapsed_time, report_elapsed_time)

    def analyze_start(self):
        self.analyze_start_time = time.time()

    def analyze_end(self):
        self.analyze_end_time = time.time()
        self.analyze_elapsed_time = self.analyze_end_time - self.analyze_start_time

    def report_start(self):
        self.report_start_time = time.time()

    def report_end(self):
        self.report_end_time = time.time()
        self.report_elapsed_time = self.report_end_time - self.report_start_time

#------------------------------------------------------------------------------

class AnalyzeStatistics(object):
    def __init__(self, num_plugins):
        self.num_plugins = num_plugins
        self.cur_plugin = None
        self.called_plugins = []
        self.start_time = None
        self.end_time = None
        self.elapsed_time = None

    def __str__(self):
        elapsed_time = None
        avg_plugin_time = None
        n_called = len(self.called_plugins)
        if self.elapsed_time is not None:
            elapsed_time = format_elapsed_time(self.elapsed_time)
            if n_called:
                avg_plugin_time = format_elapsed_time(self.elapsed_time / n_called)

        return "%d/%d plugins in %s elapsed, avg plugin %s elapsed, plugins=[\n%s\n]" % \
               (n_called, self.num_plugins, elapsed_time, avg_plugin_time,
                self.called_plugins_to_string())

    def called_plugins_to_string(self):
        return '\n'.join([str(x) for x in self.called_plugins])
                                               
    def start(self):
        self.start_time = time.time()

    def end(self):
        self.end_time = time.time()
        self.elapsed_time = self.end_time - self.start_time

    def new_plugin(self, plugin):
        self.cur_plugin = PluginStatistics(plugin)
        self.called_plugins.append(self.cur_plugin)
        self.cur_plugin.analyze_start()

#------------------------------------------------------------------------------

class Analyze(object):
    def __init__(self):
        self.plugins = load_plugins()
        log_avc.debug("Number of Plugins = %d", len(self.plugins))
        
    def analyze_avc(self, avc, report_receiver, get_env=True, line_numbers=None):
        log_avc.debug("analyze_avc() avc=%s", avc)
        statistics = AnalyzeStatistics(len(self.plugins))
        statistics.start()
        for plugin in self.plugins:
            statistics.new_plugin(plugin)
            statistics.cur_plugin.analyze_start()
            try:	
                plugin.prepare(avc, get_env)
                if plugin.analyze():
                    statistics.cur_plugin.analyze_end()
                    statistics.cur_plugin.report_start()
                    plugin.report(report_receiver, line_numbers)
                    statistics.cur_plugin.report_end()
                    break;

            except Exception, e:
                log_avc.exception("Plugin Exception %s " % plugin.analysis_id) 
                self.plugins.remove(plugin)

            statistics.cur_plugin.analyze_end()

        statistics.end()
        log_stats.debug("analyze_avc() avc=%s\nstatistics=%s", avc, statistics)
    
#------------------------------------------------------------------------------

class AnalyzeThread(Analyze, threading.Thread):
    def __init__(self,queue):
        # parent class constructors
        threading.Thread.__init__(self)
        Analyze.__init__(self)

        self.queue=queue
        
    def run(self):
        while True:
            avc, report_receiver = self.queue.get()
            self.analyze_avc(avc, report_receiver)

#------------------------------------------------------------------------------

class PluginReportReceiver(object):
    def __init__(self, database):
        self.database = database

    def report_problem(self, siginfo):
        timestamp = TimeStamp()
        try:
            database_siginfo = self.database.lookup_signature(siginfo.sig)
            database_siginfo.update_merge(siginfo)
            log_database.debug("signature found in database")
        except ProgramError, e:
            if e.errno == ERR_NO_SIGNATURE_MATCH:
                log_database.debug("not in database yet")
                siginfo.local_id=self.database.sigs.generate_local_id()
                siginfo.first_seen_date = str(timestamp)
                database_siginfo = self.database.add_siginfo(siginfo)
            else:
                raise

        database_siginfo.last_seen_date = str(timestamp)
        database_siginfo.report_count += 1
        self.database.save()
        return database_siginfo


#------------------------------------------------------------------------------

class SETroubleshootDatabase:
    def __init__(self, name):
        self.database_dir = cfg.get('database','database_dir')
        # strip off extension if one was provided
        self.name = re.sub('\\.xml$', '', name)
        self.filename = self.name + '.xml'
        self.filepath = os.path.join(self.database_dir, self.filename)
        self.lock = threading.Lock()
        self.load()

    def load(self):
        self.sigs = SEFaultSignatureSet()
        
        if os.path.exists(self.filepath):
           stat_info = os.stat(self.filepath)
           if stat_info[ST_SIZE] > 0:
               self.sigs.read_xml_file(self.filepath, 'sigs')

    def save(self):
        log_database.debug("writing database (%s)", self.filepath)
        self.sigs.write_xml('sigs', self.filepath)
        
    def acquire(self):
        self.lock.acquire()

    def release(self):
        self.lock.release()
    
    def lookup_signature(self, sig):
        siginfo = None

        matches = self.sigs.match_signatures(sig)
        log_sig.debug("lookup_signature: found %d matches with scores %s",
                      len(matches), ",".join(["%.2f" % x.score for x in matches]))
        if len(matches) == 0:
            raise ProgramError(ERR_NO_SIGNATURE_MATCH)
        elif len(matches) > 1:
            raise ProgramError(ERR_MULTIPLE_SIGNATURE_MATCH)
        siginfo = matches[0].siginfo
        return siginfo

    def lookup_local_id(self, local_id):
        return self.sigs.lookup_local_id(local_id)

    def add_siginfo(self, siginfo):
        return self.sigs.add_siginfo(siginfo)

    def remove_siginfo(self, siginfo):
        self.sigs.remove_siginfo(siginfo)

    def get_sigs_filtered(self, criteria):
        return self.sigs

#------------------------------------------------------------------------------

def map_logfile_to_database_name(logfile_path):
    logfile_basename = os.path.basename(logfile_path)
    return logfile_basename

def analyze_logfile(logfile_path):
    analyze = Analyze()

    def avc_event_handler(event):
        log_avc.debug('avc_event_handler() event=%s', event)
        if event.is_avc and not event.is_granted and len(event.audit_dict) > 0:
            avc=AVC(event.audit_dict)
            analyze.analyze_avc(avc, report_receiver, False, event.line_numbers.keys())

    def progress_callback(progress):
        print "\r%3d%% done" % (progress*100)
        log_avc.debug("%s: %d%% done", logfile_path, progress*100)

    log_avc.debug('analyze_logfile: file=%s', logfile_path)

    database_filename = map_logfile_to_database_name(logfile_path)
    if os.path.exists(database_filename):
        os.remove(database_filename)

    logfile_database = SETroubleshootDatabase(database_filename)
    report_receiver = PluginReportReceiver(logfile_database)
    sigs, statistics = parse_audit_logfile(logfile_path, avc_event_handler, report_receiver, progress_callback)
    #logfile_database.save()

    log_stats.debug('analyze_logfile: stats=%s', statistics)
    #log_avc.debug('analyze_logfile: results=%s', sigs)

    seperator = '-'*80 + '\n'
    print "found %d alerts in %s" % (len(sigs.signature_list), logfile_path)
    print seperator
    for siginfo in sigs.signature_list:
        print "=== alert composed from data on lines: %s ===" % ','.join([str(x) for x in siginfo.line_numbers])
        print siginfo.format_text()
        print seperator
        

    return None
        
